@@ -1,4 +1,4 @@
-#!/usr/bin/env perl
+#!perl
 $|++;
 
 use lib qw(./lib);
@@ -25,6 +25,8 @@ my $opt_skip      = undef;
 my $opt_real      = undef;
 my $opt_keep      = undef;
 my $opt_delete    = undef;
+my $opt_force     = undef;
+my $opt_etag      = undef;
 my $opt_log       = 'log4perl.conf';
 my $opt_header    = [];
 
@@ -35,6 +37,8 @@ GetOptions(
     "skip"         => \$opt_skip ,
     "keep"         => \$opt_keep ,
     "delete"       => \$opt_delete ,
+    "etag=s"       => \$opt_etag ,
+    "force|f"      => \$opt_force ,
     "r"            => \$opt_recursive ,
     "x"            => \$opt_real ,
     "H=s@"         => \$opt_header ,
@@ -70,6 +74,9 @@ elsif ($cmd eq 'put') {
 elsif ($cmd eq 'post') {
     $ret = cmd_post(@ARGV);
 }
+elsif ($cmd eq 'patch') {
+    $ret = cmd_patch(@ARGV);
+}
 elsif ($cmd eq 'delete') {
     $ret = cmd_delete(@ARGV);
 }
@@ -132,6 +139,7 @@ usage: $0 [options] get /path | url
 usage: $0 [options] put (/path/ | url)       # create a folder 
 usage: $0 [options] put (/path | url) file [mimeType]
 usage: $0 [options] post (/path | url) file [mimeType]
+usage: $0 [options] patch (/path | url) (file | SPARQL)
 usage: $0 [options] head /path | url
 usage: $0 [options] options /path | url
 usage: $0 [options] delete /path | url
@@ -147,6 +155,7 @@ options:
     --skip                   - skip files that already exist (mirror)
     --delete                 - delete local files that are not at the remote location (mirror)
     --keep                   - keep containers (clean)
+    --force|f                - force overwriting existing resources (put,patch)
     -r                       - recursive (mirror, upload, clean)
     -x                       - do it for real (upload, clean)
     -H "Header"              - add a header (get,post,put,head,delete)
@@ -200,8 +209,10 @@ sub _cmd_list {
 prefix ldp: <http://www.w3.org/ns/ldp#> 
 
 SELECT ?resource ?type {
-    ?resource a ?type .
-    FILTER (?type IN (
+    ?base ldp:contains ?resource .
+    OPTIONAL {
+        ?resource a ?type .
+        FILTER (?type IN (
                       ldp:Resource,
                       ldp:RDFSource,
                       ldp:Container,
@@ -210,6 +221,7 @@ SELECT ?resource ?type {
                       ldp:NonRDFSource
                       ) 
             )
+    }
 }
 EOF
 
@@ -218,15 +230,30 @@ EOF
     $util->sparql($model, $sparql, sub {
         my $res = shift;
         my $name = $res->value('resource')->as_string; 
-        $name = substr($name,length($webbase));
-        $name =~ s/^\///; 
-        my $type = $res->value('type')->as_string;
 
-        if ($FILES{$url . $name} && $FILES{$url . $name} eq 'container') {
+        # clean absolute urls into relative ones ...
+        if ($name =~ /^http/) {
+            $name = substr($name,length($iri));
+        }
+        $name =~ s/^\///; 
+
+        # Read the type from type or guess ..based on the name :P
+        my $type;
+
+        if ($res->value('type')) {
+            $type = $res->value('type')->as_string;
+        }
+        else {
+            $type = ($name =~ /\/$/) ? 'Container'  : 'Resource';
+        }
+
+        my $key = $url . $name;
+
+        if (exists $FILES{$key} && $FILES{$key} eq 'container') {
             # Containers are more interesting than resources
         }
         else {
-            $FILES{$url . $name} = $type =~ /Container/ ? "container" : "resource";
+            $FILES{$key} = $type =~ /Container/ ? "container" : "resource";
         }
     });
 
@@ -347,7 +374,20 @@ sub cmd_put {
 
     my $iri = _make_url($url);
     my %headers = _make_headers();
-    $headers{'Content-Type'} = $mimeType unless $headers{'Content-Type'};
+
+    if ($mimeType) {
+        $headers{'Content-Type'} = $mimeType unless $headers{'Content-Type'};
+    }
+
+    # Prevent overwriting exiting resources    
+    unless ($opt_force) {
+        $headers{'If-None-Match'} = '*';
+    }
+
+    # Prevent overwriting changed resources
+    if ($opt_etag) {
+        $headers{'If-Match'} = $opt_etag;
+    }
 
     my $response;
 
@@ -355,6 +395,7 @@ sub cmd_put {
         $response = $agent->put($iri, $data, %headers);
     }
     else {
+        %headers = _link_header('<http://www.w3.org/ns/ldp#BasicContainer>; rel="type"', %headers);
         $response = $agent->put($iri,undef,%headers);
     }
 
@@ -383,6 +424,7 @@ sub cmd_post {
 
     my $iri = _make_url($url);
     my %headers = _make_headers();
+
     $headers{'Content-Type'} = $mimeType unless $headers{'Content-Type'}; 
 
     my $response = $agent->post($iri, $data, %headers);
@@ -400,6 +442,52 @@ sub cmd_post {
     return 0;
 }
 
+sub cmd_patch {
+    my ($url,$file_or_string) = @_;    
+
+    unless ($url && $file_or_string) {
+        print STDERR "Need a url and a file or string\n";
+        return 1;
+    }
+
+    my $sparql;
+
+    $file_or_string = '/dev/stdin' if $file_or_string eq '-';
+
+    if (-r $file_or_string) {
+        $sparql = path($file_or_string)->slurp_utf8;
+    }
+    else {
+        $sparql = $file_or_string;
+    }
+
+    my $iri = _make_url($url);
+    my %headers = _make_headers();
+    $headers{'Content-Type'} = 'application/sparql-update' unless $headers{'Content-Type'}; 
+
+    # Prevent overwriting exiting resources    
+    unless ($opt_force) {
+        $headers{'If-None-Match'} = '*';
+    }
+
+    # Prevent overwriting changed resources
+    if ($opt_etag) {
+        $headers{'If-Match'} = $opt_etag;
+    }
+
+    my $response = $agent->patch($iri, $sparql, %headers);
+
+    unless ($response->is_success) {
+        printf STDERR "%s - failed to $url\n" , $response->code;
+        printf STDERR "%s\n" , $response->message;
+        return 2;
+    }
+
+    print STDERR $response->decoded_content , "\n";
+
+    return 0;
+}
+
 sub cmd_delete {
     my ($url) = @_;
 
@@ -667,7 +755,7 @@ sub cmd_headers {
 
     usage() unless $method && $url;
 
-    my $headers = _headers($method,$url);
+    my $headers = _authentication_headers($method,$url);
 
     print "$headers\n";
 
@@ -693,7 +781,7 @@ sub cmd_curl {
         @rest = map { String::Escape::quote($_) } @rest;
     }
 
-    my $headers = _headers($method,$url);
+    my $headers = _authentication_headers($method,$url);
     my $opts    = join(" ",@rest);
     system("curl $headers $opts") == 0;
 }
@@ -759,6 +847,7 @@ sub cmd_id_token {
     return 0;
 }
 
+# Parse the optional provided headers into a hash
 sub _make_headers {
     my %headers = ();
     for my $h (@$opt_header) {
@@ -768,6 +857,21 @@ sub _make_headers {
     return %headers;
 }
 
+# Add a Link header 
+sub _link_header {
+    my ($link,%headers) = @_;
+
+    if (exists $headers{Link}) {
+        $headers{Link} .= ", $link";
+    }
+    else {
+        $headers{Link} = $link;
+    }
+
+    %headers;
+}
+
+# Expand a relative url to a full url with SOLID_REMOTE_BASE
 sub _make_url {
     my $url = shift;
 
@@ -780,7 +884,8 @@ sub _make_url {
     return "$webbase$1";
 }
 
-sub _headers {
+# Expand the CURL header with authentication and DPop headers
+sub _authentication_headers {
     my ($method,$url) = @_;
 
     $webid //= $url;
@@ -799,6 +904,7 @@ sub _headers {
     return join(" ",@headers);
 }
 
+# Guess the mime type of a file
 sub _guess_mimetype {
     my ($path) = @_;
     
@@ -870,10 +976,15 @@ solid_auth.pl - A Solid management tool
       solid_auth.pl post /inbox/ myfile.jsonld 
 
       # Put some data
-      solid_auth.pl put /public/myfile.txt myfile.txt 
+      solid_auth.pl -f put /public/myfile.txt myfile.txt 
 
+      # Patch data
+      solid_auth.pl -f patch /public/myfile.txt.meta  - <<EOF
+      INSERT DATA { <> <http://example.org> 1234 }
+      EOF
+      
       # Create a folder
-      solid_auth.pl put /public/mytestfolder/
+      solid_auth.pl -f put /public/mytestfolder/
 
       # Delete some data
       solid_auth.pl delete /public/myfile.txt
@@ -944,6 +1055,26 @@ Delete local files that are not in the remote container (mirror).
 
 Keep containers when cleaning data (clean).
 
+=item --etag=STRING
+
+Only update the data when the 'Etag' header matches the given string (put,patch). E.g.
+use the C<head> command to find the ETag of a resource :
+
+    $ solid_auth.pl head /demo/LICENSE
+    ...
+    ETag: "189aa19989dc47eab46c9f2e8c47d0836bb08cb09f7863cbf3cd3bb9a751be27"
+    ...
+
+Now update the resource with ETag protection
+
+    $ solid_auth.pl \
+        --etag=189aa19989dc47eab46c9f2e8c47d0836bb08cb09f7863cbf3cd3bb9a751be27 \
+        put /demo/LICENSE LICENSE
+
+=item --force | -f
+
+Force overwriting existing resources (put, patch).
+
 =item -r
 
 Recursive (clean, mirror, upload).
@@ -1016,6 +1147,11 @@ Uses libmagic to guess the mimetype.
 
 When the URL ends with a slash (/), then a new container will be created.
 
+=item patch URL FILE|SPARQL
+
+Send the contents of a SPARQL patch file or string to a URL. Return the
+HTTP Message of the HTTP PATCH request to the URL.
+
 =item head URL
 
 Return the HTTP Header of a HTTP HEAD request to URL.
